;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "dap-mode" "20260523.830"
  "Debug Adapter Protocol mode."
  '((emacs        "29.1")
    (dash         "2.18.0")
    (lsp-mode     "6.0")
    (bui          "1.1.0")
    (f            "0.20.0")
    (s            "1.12.0")
    (lsp-treemacs "0.1")
    (posframe     "0.7.0")
    (ht           "2.3")
    (lsp-docker   "1.0.0"))
  :url "https://github.com/emacs-lsp/dap-mode"
  :commit "fc78b2a1db5f30e65875653b4807b801c0ef23bc"
  :revdesc "fc78b2a1db5f"
  :keywords '("languages" "debug")
  :authors '(("Ivan Yonchovski" . "yyoncho@gmail.com"))
  :maintainers '(("Ivan Yonchovski" . "yyoncho@gmail.com")))
